unit Unit1;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls;

type
  PrintAttribute = class( TCustomAttribute )
  protected
    FFormatSttring  : String;

  public
    property FormatSttring : String read FFormatSttring;

  public
    constructor Create( AFormatSttring : String = '' );

  end;

  CaptionAttribute = class( TCustomAttribute )
  protected
    FCaption  : String;

  public
    property Caption : String read FCaption;

  public
    constructor Create( ACaption : String );

  end;

  TAddress = class
  protected
    FNumber   : Integer;
    FStreat   : String;
    FCity     : String;
    FZipCode  : Integer;

  public
    constructor Create(
        ANumber : Integer;
        AStreat : String;
        ACity : String;
        AZipCode : Integer
      );

  public
    [Print]
    property Number   : Integer read FNumber;

    [Print]
    [Caption( 'Streat Name' )]
    property Streat   : String  read FStreat;

    [Print]
    property City     : String  read FCity;

    [Print]
    property ZipCode  : Integer read FZipCode;

  end;

  TAccount = class
  protected
    FUserName   : String;
    FFirstName  : String;
    FLastName   : String;

    FAge        : TDate;

    FAddress    : TAddress;

    FAmount     : Single;

  public
    constructor Create(
        AUserName : String;
        AFirstName : String;
        ALastName : String;
        AAge : TDate;
        AAmount : Single;
        ANumber : Integer;
        AStreat : String;
        ACity : String;
        AZipCode : Integer
      );

    destructor  Destroy(); override;

  public
    property UserName   : String    read FUserName;

    [Print]
    [Caption( 'First Name' )]
    property FirstName  : String    read FFirstName;

    [Print]
    [Caption( 'Last Name' )]
    property LastName   : String    read FLastName;

    [Print( 'dd/mm/yyyy' )]
    property Age        : TDate     read FAge;

    [Print]
    property Address    : TAddress  read FAddress;

    [Print( '%8.2f' )]
    [Caption( 'Available Amount' )]
    property Amount     : Single    read FAmount;

  end;

  TForm1 = class(TForm)
    Memo1: TMemo;
    Button1: TButton;
    procedure Button1Click(Sender: TObject);
  private
    { Private declarations }
//    FList<>
  public
    { Public declarations }
    procedure Print( AObject : TObject );

  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

uses
  Mitov.TypeInfo, System.DateUtils;

constructor PrintAttribute.Create( AFormatSttring : String = '' );
begin
  inherited Create();
  FFormatSttring := AFormatSttring;
end;

constructor CaptionAttribute.Create( ACaption : String );
begin
  inherited Create();
  FCaption := ACaption;
end;

procedure TForm1.Button1Click(Sender: TObject);
var
  AAccount : TAccount;

begin
  AAccount := TAccount.Create(
      'jsmith',
      'John',
      'Smith',
      IncYear( Now(), - 30 ),
      1025.35,

      4735,
      'South Streat',
      'Codeville',
      81295
    );

  Memo1.Lines.Clear();
  Print( AAccount );
end;

procedure TForm1.Print( AObject : TObject );
begin
  AObject.TypeInfo.SingleProperties.Query().ForEach(
      procedure( AItem : ISinglePropertyInfo )
      var
        ACaption : String;

      begin
        ACaption := AItem.Name;
        AItem.AccessAttributes.ForEach<CaptionAttribute>(
            procedure( ACaptionAttribute : CaptionAttribute )
            begin
              ACaption := ACaptionAttribute.Caption;
            end
          );

        ACaption := ACaption + ' : ';

        AItem.AccessAttributes.ForEach<PrintAttribute>(
            procedure( APrintAttribute : PrintAttribute )
            var
              AFormatString : String;

            begin
              AFormatString := APrintAttribute.FormatSttring;
              if( AItem.TypeInfo.InheritsFrom( TRttiInfo.TypeOf<TObject>()) ) then
                Print( AItem.Value[ AObject ].AsObject )

              else if( AItem.TypeInfo = TRttiInfo.TypeOf<Single>() ) then
                begin
                if( AFormatString <> '' ) then
                  Memo1.Lines.Add( ACaption + String.Format( AFormatString, [ AItem.Value[ AObject ].AsType<Single> ] ))

                else
                  Memo1.Lines.Add( ACaption + AItem.Value[ AObject ].AsType<Single>.ToString() );

                end

              else if( AItem.TypeInfo = TRttiInfo.TypeOf<Integer>() ) then
                begin
                if( AFormatString <> '' ) then
                  Memo1.Lines.Add( ACaption + String.Format( AFormatString, [ AItem.Value[ AObject ].AsType<Integer> ] ))

                else
                  Memo1.Lines.Add( ACaption + AItem.Value[ AObject ].AsType<Integer>.ToString() );

                end

              else if( AItem.TypeInfo = TRttiInfo.TypeOf<String>() ) then
                begin
                if( AFormatString <> '' ) then
                  Memo1.Lines.Add( ACaption + String.Format( AFormatString, [ AItem.Value[ AObject ].AsString ] ))

                else
                  Memo1.Lines.Add( ACaption + AItem.Value[ AObject ].AsString() );

                end

              else if( AItem.TypeInfo = TRttiInfo.TypeOf<TDate>() ) then
                begin
                if( AFormatString <> '' ) then
                  Memo1.Lines.Add( ACaption + FormatDateTime( AFormatString, AItem.Value[ AObject ].AsType<TDate> ) )

                else
                  Memo1.Lines.Add( ACaption + DateToStr( AItem.Value[ AObject ].AsType<TDate> ));

                end

            end
          );

      end
    );

end;

constructor TAccount.Create(
        AUserName : String;
        AFirstName : String;
        ALastName : String;
        AAge : TDate;
        AAmount : Single;
        ANumber : Integer;
        AStreat : String;
        ACity : String;
        AZipCode : Integer
      );
begin
  inherited Create();

  FUserName := AUserName;
  FFirstName := AFirstName;
  FLastName := ALastName;
  FAge := AAge;
  FAmount := AAmount;

  FAddress := TAddress.Create(
      ANumber,
      AStreat,
      ACity,
      AZipCode
    );

end;

destructor TAccount.Destroy();
begin
  FAddress.Free();
  inherited;
end;

constructor TAddress.Create(
        ANumber : Integer;
        AStreat : String;
        ACity : String;
        AZipCode : Integer
      );
begin
  inherited Create();

  FNumber := ANumber;
  FStreat := AStreat;
  FCity := ACity;
  FZipCode := AZipCode
end;

end.
