unit Unit1;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls;

type
  TMyNameAttribute = class( TCustomAttribute )
  protected
    FName : String;

  public
    constructor Create( AName : String );

  public
    property Name : String read FName;

  end;

  TMyFiledAttribute = class( TCustomAttribute );

  [TMyFiled]
  TMyFieldData = class( TPersistent )
  end;

  [TMyName( 'Hello' )]
  TMyObject = class( TObject )
  protected
    [TMyFiled]
    FField1 : Integer;

    FField2 : Integer;

    FData   : TMyFieldData;

  end;

  [TMyName( 'Hello 1' )]
  TMyObject1 = class( TMyObject )
  protected
    [TMyFiled]
    FField3 : String;

    FField4 : String;

  end;

  TMyObject2 = class( TMyObject1 )
  protected
    FField5 : Real;

  end;

  TForm1 = class(TForm)
    Memo1: TMemo;
    GroupBox1: TGroupBox;
    Button1: TButton;
    Button3: TButton;
    GroupBox2: TGroupBox;
    Button2: TButton;
    Button4: TButton;
    GroupBox3: TGroupBox;
    Button5: TButton;
    Button6: TButton;
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure Button4Click(Sender: TObject);
    procedure Button5Click(Sender: TObject);
    procedure Button6Click(Sender: TObject);
  private
    { Private declarations }
    AObject2 : TMyObject;

  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

uses
  System.Rtti, Mitov.TypeInfo;

procedure TForm1.Button1Click(Sender: TObject);
var
  AContext  : TRttiContext;

  procedure ProcessTypeAttributes( AType : TRttiType );
  var
    AAttribute  : TCustomAttribute;

  begin
    for AAttribute in AType.GetAttributes() do
      if( AAttribute is TMyNameAttribute ) then
        begin
        Memo1.Lines.Add( TMyNameAttribute( AAttribute ).Name );
        Exit;
        end;

    if( AType.BaseType <> NIL ) then
      ProcessTypeAttributes( AType.BaseType );

  end;

begin
  ProcessTypeAttributes( AContext.GetType( AObject2.ClassType() ));
end;

constructor TMyNameAttribute.Create( AName : String );
begin
  inherited Create();
  FName := AName;
end;

procedure TForm1.Button2Click(Sender: TObject);
var
  ANameAttribute  : TMyNameAttribute;

begin
  if( AObject2.TypeInfo().AccessAttributes.Get<TMyNameAttribute>( ANameAttribute )) then
    Memo1.Lines.Add( ANameAttribute.Name );

end;

procedure TForm1.Button3Click(Sender: TObject);
var
  AContext  : TRttiContext;

  procedure ProcessTypeAttributes( AType : TRttiType; AFieldName : String  );
  var
    AAttribute  : TCustomAttribute;

  begin
    for AAttribute in AType.GetAttributes() do
      if( AAttribute is TMyFiledAttribute ) then
        begin
        Memo1.Lines.Add( AFieldName );
        Exit;
        end;

    if( AType.BaseType <> NIL ) then
      ProcessTypeAttributes( AType.BaseType, AFieldName );

  end;

var
  AField          : TRttiField;
  AAttribute      : TCustomAttribute;
  AAttributeFound : Boolean;

begin
  for AField in AContext.GetType( AObject2.ClassType() ).GetFields() do
    begin
    AAttributeFound := False;
    for AAttribute in AField.GetAttributes() do
      if( AAttribute is TMyFiledAttribute ) then
        begin
        Memo1.Lines.Add( AField.Name );
        AAttributeFound := True;
        Break;
        end;

    if( not AAttributeFound ) then
      ProcessTypeAttributes( AField.FieldType, AField.Name );

    end;

end;

procedure TForm1.Button4Click(Sender: TObject);
var
  AField : IFieldInfo;

begin
  for AField in AObject2.TypeInfo().Fields do
    if( AField.HasCustomAttribute( TMyFiledAttribute, True )) then
      Memo1.Lines.Add( AField.Name );

end;

procedure TForm1.Button5Click(Sender: TObject);
begin
  AObject2.TypeInfo().AccessAttributes.ForLast<TMyNameAttribute>(
      procedure( ANameAttribute : TMyNameAttribute )
      begin
        Memo1.Lines.Add( ANameAttribute.Name );
      end
    );

end;

procedure TForm1.Button6Click(Sender: TObject);
begin
  AObject2.TypeInfo().GetFields(
      function( AField : IFieldInfo ) : Boolean
      begin
        Result := AField.HasCustomAttribute( TMyFiledAttribute, True );
      end

    ).Query().ForEach(
        procedure( AField : IFieldInfo )
        begin
          Memo1.Lines.Add( AField.Name );
        end
      );


end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  AObject2 := TMyObject2.Create();
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  AObject2.Free();
end;

end.


